import datetime

from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.authentication import BasicAuthAuthenticationPolicy
from pyramid.config import Configurator
from pyramid.renderers import JSON

from channelstream.ext_json import json
from channelstream.wsgi_views.wsgi_security import APIFactory


def datetime_adapter(obj, request):
    return obj.isoformat()


def make_app(server_config):
    config = Configurator(settings=server_config, root_factory=APIFactory)

    def check_function(username, password, request):
        if password == server_config['admin_secret']:
            return 'admin'
        return None

    authn_policy = BasicAuthAuthenticationPolicy(check_function,
                                                 realm='ChannelStream')
    authz_policy = ACLAuthorizationPolicy()
    config.set_authentication_policy(authn_policy)
    config.set_authorization_policy(authz_policy)
    json_renderer = JSON(serializer=json.dumps, indent=4)
    json_renderer.add_adapter(datetime.datetime, datetime_adapter)
    config.add_renderer('json', json_renderer)
    config.add_static_view('static', path='channelstream:static/')
    config.include('pyramid_jinja2')
    config.include('channelstream.wsgi_views')
    config.scan('channelstream.wsgi_views')
    app = config.make_wsgi_app()
    return app



import logging

from datetime import datetime, timedelta

import gevent
import six

import channelstream

log = logging.getLogger(__name__)


def gc_conns():
    with channelstream.lock:
        start_time = datetime.utcnow()
        threshold = start_time - timedelta(seconds=15)
        collected_conns = []
        # collect every ref in chanels
        # remove connections from channels
        for channel in six.itervalues(channelstream.CHANNELS):
            for username, conns in list(channel.connections.items()):
                for conn in conns:
                    if conn.last_active < threshold:
                        channel.connections[username].remove(conn)
                        collected_conns.append(conn)
                channel.after_parted(username)
        # remove old conns from users and connection dictionaries
        for conn in collected_conns:
            if conn.username in channelstream.USERS:
                if conn in channelstream.USERS[conn.username].connections:
                    channelstream.USERS[conn.username].connections.remove(conn)
            if conn.id in channelstream.CONNECTIONS:
                del channelstream.CONNECTIONS[conn.id]
            # make sure connection is closed after we garbage
            # collected it from our list
            if conn.socket:
                try:
                    conn.socket.close()
                except Exception:
                    raise
        log.debug('gc_conns() time %s' % (datetime.utcnow() - start_time))


def gc_users():
    with channelstream.lock:
        start_time = datetime.utcnow()
        threshold = datetime.utcnow() - timedelta(days=1)
        for user in list(six.itervalues(channelstream.USERS)):
            if user.last_active < threshold:
                channelstream.USERS.pop(user.username)
        log.debug('gc_users() time %s' % (datetime.utcnow() - start_time))


def gc_users_forever():
    try:
        gc_users()
    finally:
        gevent.spawn_later(60, gc_users_forever)


def gc_conns_forever():
    try:
        gc_conns()
    finally:
        gevent.spawn_later(1, gc_conns_forever)



import six
from itsdangerous import TimestampSigner
from pyramid.security import (Allow,
                              Everyone,
                              ALL_PERMISSIONS,
                              authenticated_userid)


class RequestBasicChallenge(Exception):
    pass


class APIFactory(object):
    def __init__(self, request):
        self.__acl__ = []
        config = request.registry.settings
        req_url_secret = request.params.get('secret')
        req_secret = request.headers.get('x-channelstream-secret',
                                         req_url_secret)

        if request.environ['REMOTE_ADDR'] not in config['allow_posting_from']:
            return
        if req_secret:
            signer = TimestampSigner(config['secret'])
            unsigned = signer.unsign(req_secret)
            if not six.PY2:
                unsigned = unsigned.decode('utf8')
            if request.path != unsigned:
                return
        else:
            return
        self.__acl__ = [(Allow, Everyone, ALL_PERMISSIONS)]


class BasicAuthFactory(object):
    def __init__(self, request):
        self.__acl__ = []
        user = authenticated_userid(request)
        if user != 'admin':
            raise RequestBasicChallenge()
        self.__acl__ = [(Allow, Everyone, ALL_PERMISSIONS)]



import logging
import uuid
from datetime import datetime
import gevent
import six
from gevent.queue import Queue, Empty
from pyramid.view import view_config
from pyramid.httpexceptions import HTTPUnauthorized
from pyramid.security import forget
import channelstream
from channelstream.user import User
from channelstream.connection import Connection
from channelstream.channel import Channel
from ..ext_json import json

log = logging.getLogger(__name__)


def pass_message(msg, stats):
    if msg.get('timestamp'):
        # if present lets use timestamp provided in the message
        if '.' in msg['timestamp']:
            timestmp = datetime.strptime(msg['timestamp'],
                                         '%Y-%m-%dT%H:%M:%S.%f')
        else:
            timestmp = datetime.strptime(msg['timestamp'],
                                         '%Y-%m-%dT%H:%M:%S')
    else:
        timestmp = datetime.utcnow()
    message = {'uuid': str(uuid.uuid4()).replace('-', ''),
               'user': msg.get('user'),
               'message': msg['message'],
               'type': 'message',
               'timestamp': timestmp}
    pm_users = msg.get('pm_users', [])
    total_sent = 0
    stats['total_unique_messages'] += 1
    if msg.get('channel'):
        channel_inst = channelstream.CHANNELS.get(msg['channel'])
        if channel_inst:
            total_sent += channel_inst.add_message(message,
                                                   pm_users=pm_users)
    elif pm_users:
        # if pm then iterate over all users and notify about new message!
        for username in pm_users:
            user_inst = channelstream.USERS.get(username)
            if user_inst:
                total_sent += user_inst.add_message(message)
    stats['total_messages'] += total_sent


def get_connection_channels(connection):
    found_channels = []
    for channel in six.itervalues(channelstream.CHANNELS):
        user_conns = channel.connections.get(connection.username) or []
        if connection in user_conns:
            found_channels.append(channel.name)
    return found_channels


class ServerViews(object):
    def __init__(self, request):
        self.request = request

    def _get_channel_info(self, req_channels=None, include_history=True,
                          include_connections=False, include_users=False,
                          exclude_channels=None):
        """
        Gets channel information for req_channels or all channels
        if req_channels is not present
        :param: include_history (bool) will include message history
                for the channel
        :param: include_connections (bool) will include connection list
                for the channel
        :param: include_users (bool) will include user list for the channel
        :param: exclude_channels (bool) will exclude specific channels
                from info list (handy to exclude global broadcast)
        """

        if not exclude_channels:
            exclude_channels = []
        start_time = datetime.utcnow()

        json_data = {"channels": {},
                     "unique_users": len(channelstream.USERS),
                     "users": []}

        users_to_list = []

        # select everything for empty list
        if not req_channels:
            channel_instances = six.itervalues(channelstream.CHANNELS)
        else:
            channel_instances = [channelstream.CHANNELS[c]
                                 for c in req_channels]

        for channel_inst in channel_instances:
            if channel_inst.name in exclude_channels:
                continue

            json_data["channels"][channel_inst.name] = {'history': []}
            chan_info = json_data["channels"][channel_inst.name]
            if include_history:
                chan_info['history'] = channel_inst.history
            chan_info['total_users'] = len(channel_inst.connections)
            chan_info['total_connections'] = sum(
                [len(conns) for conns in channel_inst.connections.values()])
            chan_info['users'] = []
            for username in channel_inst.connections.keys():
                user_inst = channelstream.USERS.get(username)
                if include_users and user_inst.username not in users_to_list:
                    users_to_list.append(user_inst.username)
                udata = {'user': user_inst.username,
                         "connections": []}
                if include_connections:
                    udata['connections'] = [conn.id for conn in
                                            channel_inst.connections[username]]
                chan_info['users'].append(udata)
            chan_info['last_active'] = channel_inst.last_active

        for username in users_to_list:
            json_data['users'].append(
                {'user': username,
                 'state': channelstream.USERS[username].state})
        log.info('info time: %s' % (datetime.utcnow() - start_time))
        return json_data

    def get_common_info(self, channels, info_config):
        include_history = info_config.get('include_history', True)
        include_users = info_config.get('include_users', True)
        exclude_channels = info_config.get('exclude_channels', [])
        include_connections = info_config.get('include_connections', False)
        channels_info = self._get_channel_info(
            channels, include_history=include_history,
            include_connections=include_connections,
            include_users=include_users, exclude_channels=exclude_channels)
        return channels_info

    @view_config(route_name='action', match_param='action=connect',
                 renderer='json', permission='access')
    def connect(self):
        """
        return the id of connected users - will be secured with password string
        for webapp to internally call the server - we combine conn string
        with user id, and we tell which channels the user is allowed to
        subscribe to
        """
        json_body = self.request.json_body
        username = json_body.get('username')
        fresh_user_state = json_body.get('fresh_user_state', {})
        update_user_state = json_body.get('user_state', {})
        channel_configs = json_body.get('channel_configs', {})
        state_public_keys = json_body.get('state_public_keys', None)
        conn_id = json_body.get('conn_id')
        subscribe_to_channels = json_body.get('channels')
        if username is None:
            self.request.response.status = 400
            return {'error': "No username specified"}
        if not subscribe_to_channels:
            self.request.response.status = 400
            return {'error': "No channels specified"}

        # everything is ok so lets add new connection to
        # channel and connection list
        with channelstream.lock:
            if username not in channelstream.USERS:
                user = User(username)
                user.state_from_dict(fresh_user_state)
                channelstream.USERS[username] = user
            else:
                user = channelstream.USERS[username]
            if state_public_keys is not None:
                user.state_public_keys = state_public_keys

            user.state_from_dict(update_user_state)
            connection = Connection(username, conn_id)
            if connection.id not in channelstream.CONNECTIONS:
                channelstream.CONNECTIONS[connection.id] = connection
            user.add_connection(connection)
            for channel_name in subscribe_to_channels:
                # user gets assigned to a channel
                if channel_name not in channelstream.CHANNELS:
                    channel = Channel(channel_name,
                                      channel_configs=channel_configs)
                    channelstream.CHANNELS[channel_name] = channel
                channelstream.CHANNELS[channel_name].add_connection(connection)
            log.info('connecting %s with uuid %s' % (username, connection.id))

        # get info config for channel information
        info_config = json_body.get('info') or {}
        channels_info = self.get_common_info(subscribe_to_channels,
                                             info_config)
        return {'conn_id': connection.id, 'state': user.state,
                'channels': subscribe_to_channels,
                'channels_info': channels_info}

    @view_config(route_name='action', match_param='action=subscribe',
                 renderer='json', permission='access')
    def subscribe(self, *args):
        """ call this to subscribe specific connection to new channels """
        json_body = self.request.json_body
        conn_id = json_body.get('conn_id', self.request.GET.get('conn_id'))
        connection = channelstream.CONNECTIONS.get(conn_id)
        subscribe_to_channels = json_body.get('channels')
        channel_configs = json_body.get('channel_configs', {})
        if not connection:
            self.request.response.status = 403
            return {'error': "Unknown connection"}
        if not subscribe_to_channels:
            self.request.response.status = 400
            return {'error': "No channels specified"}
        # everything is ok so lets add new connection to channel
        # and connection list
        # lets lock it just in case
        # find the right user
        user = channelstream.USERS.get(connection.username)
        with channelstream.lock:
            if user:
                for channel_name in subscribe_to_channels:
                    if channel_name not in channelstream.CHANNELS:
                        channel = Channel(channel_name,
                                          channel_configs=channel_configs)
                        channelstream.CHANNELS[channel_name] = channel
                    channelstream.CHANNELS[channel_name].add_connection(
                        connection)

        # get info config for channel information
        current_channels = get_connection_channels(connection)
        info_config = json_body.get('info') or {}
        channels_info = self.get_common_info(current_channels, info_config)
        return {"channels": current_channels,
                "channels_info": channels_info}

    @view_config(route_name='action', match_param='action=unsubscribe',
                 renderer='json', permission='access')
    def unsubscribe(self, *args):
        """ call this to unsubscribe specific connection from channels """
        json_body = self.request.json_body
        conn_id = json_body.get('conn_id', self.request.GET.get('conn_id'))
        connection = channelstream.CONNECTIONS.get(conn_id)
        unsubscribe_channels = json_body.get('channels')
        if not connection:
            self.request.response.status = 403
            return {'error': "Unknown connection"}
        if not unsubscribe_channels:
            self.request.response.status = 400
            return {'error': "No channels specified"}
        # everything is ok so lets add new connection to channel
        # and connection list
        # lets lock it just in case
        # find the right user
        user = channelstream.USERS.get(connection.username)
        with channelstream.lock:
            if user:
                for channel_name in unsubscribe_channels:
                    channelstream.CHANNELS[channel_name].remove_connection(
                        connection)

        # get info config for channel information
        current_channels = get_connection_channels(connection)
        info_config = json_body.get('info') or {}
        channels_info = self.get_common_info(current_channels, info_config)

        return {"channels": current_channels,
                "channels_info": channels_info}

    def _add_CORS(self):
        self.request.response.headers.add('Access-Control-Allow-Origin', '*')
        self.request.response.headers.add('XDomainRequestAllowed', '1')
        self.request.response.headers.add('Access-Control-Allow-Methods',
                                          'GET, POST, OPTIONS, PUT')
        self.request.response.headers.add('Access-Control-Allow-Headers',
                                          'Content-Type, Depth, User-Agent, '
                                          'X-File-Size, X-Requested-With, '
                                          'If-Modified-Since, X-File-Name, '
                                          'Cache-Control, Pragma, Origin, '
                                          'Connection, Referer, Cookie')
        self.request.response.headers.add('Access-Control-Max-Age', '86400')
        # self.request.response.headers.add('Access-Control-Allow-Credentials',
        #                                   'true')

    @view_config(route_name='action', match_param='action=listen',
                 request_method="OPTIONS", renderer='string')
    def handle_CORS(self):
        self._add_CORS()
        return ''

    @view_config(route_name='action', match_param='action=listen',
                 renderer='string')
    def listen(self):
        self._add_CORS()
        config = self.request.registry.settings
        self.conn_id = self.request.params.get('conn_id')
        connection = channelstream.CONNECTIONS.get(self.conn_id)
        if not connection:
            raise HTTPUnauthorized()
        # mark the conn active
        connection.last_active = datetime.utcnow()
        # attach a queue to connection
        connection.queue = Queue()

        def yield_response():
            # for chrome issues
            # yield ' ' * 1024
            # wait for this to wake up
            messages = []
            # block for first message - wake up after a while
            try:
                messages.extend(connection.queue.get(
                    timeout=config['wake_connections_after']))
            except Empty:
                pass
            # get more messages if enqueued takes up total 0.25s
            while True:
                try:
                    messages.extend(connection.queue.get(timeout=0.25))
                except Empty:
                    break
            cb = self.request.params.get('callback')
            if cb:
                yield cb + '(' + json.dumps(messages) + ')'
            else:
                yield json.dumps(messages)

        self.request.response.app_iter = yield_response()
        return self.request.response

    @view_config(route_name='action', match_param='action=user_state',
                 renderer='json', permission='access')
    def user_state(self):
        """ set the status of specific user """
        json_body = self.request.json_body
        username = json_body.get('user')
        user_state = json_body.get('user_state')
        state_public_keys = json_body.get('state_public_keys', None)
        if not username:
            self.request.response.status = 400
            return {'error': "No username specified"}

        user_inst = channelstream.USERS.get(username)
        if user_inst:
            user_inst.state_from_dict(user_state)
            if state_public_keys is not None:
                user_inst.state_public_keys = state_public_keys
            # mark active
            user_inst.last_active = datetime.utcnow()
        return user_inst.state

    @view_config(route_name='action', match_param='action=message',
                 renderer='json', permission='access')
    def message(self):
        msg_list = self.request.json_body
        for msg in msg_list:
            if not msg.get('channel') and not msg.get('pm_users', []):
                continue
            gevent.spawn(pass_message, msg, channelstream.stats)

    @view_config(route_name='action', match_param='action=disconnect',
                 renderer='json', permission='access')
    def disconnect(self):
        json_body = self.request.json_body
        conn_id = json_body.get('conn_id', self.request.GET.get('conn_id'))
        conn = channelstream.CONNECTIONS.get(conn_id)
        if conn is not None:
            conn.mark_for_gc()

    @view_config(route_name='action', match_param='action=channel_config',
                 renderer='json', permission='access')
    def channel_config(self):
        """ call this to reconfigure channels """
        json_body = self.request.json_body
        if not json_body:
            self.request.response.status = 400
            return {'error': "No channels specified"}

        with channelstream.lock:
            for channel_name, config in json_body.items():
                if not channelstream.CHANNELS.get(channel_name):
                    channel = Channel(channel_name,
                                      channel_configs=json_body)
                    channelstream.CHANNELS[channel_name] = channel
                else:
                    channel = channelstream.CHANNELS[channel_name]
                    channel.reconfigure_from_dict(json_body)
        channels_info = self._get_channel_info(json_body.keys(),
                                               include_history=False,
                                               include_users=False)
        return channels_info

    @view_config(
        context='channelstream.wsgi_views.wsgi_security:RequestBasicChallenge')
    def admin_challenge(self):
        response = HTTPUnauthorized()
        response.headers.update(forget(self.request))
        return response

    @view_config(route_name='admin',
                 renderer='templates/admin.jinja2', permission='access')
    def admin(self):
        uptime = datetime.utcnow() - channelstream.stats['started_on']
        remembered_user_count = len(
            [user for user in six.iteritems(channelstream.USERS)])
        unique_user_count = len(
            [user for user in six.itervalues(channelstream.USERS)
             if user.connections])
        total_connections = sum([len(user.connections)
             for user in six.itervalues(channelstream.USERS)])
        total_sys_connections = len(channelstream.CONNECTIONS.values())
        return {
            "remembered_user_count": remembered_user_count,
            "unique_user_count": unique_user_count,
            "total_connections": total_connections,
            "total_sys_connections": total_sys_connections,
            "total_channels": len(channelstream.CHANNELS.keys()),
            "total_messages": channelstream.stats['total_messages'],
            "total_unique_messages": channelstream.stats[
                'total_unique_messages'],
            "channels": channelstream.CHANNELS,
            "users": channelstream.USERS, "uptime": uptime
        }

    @view_config(route_name='action', match_param='action=info',
                 renderer='json', permission='access')
    def info(self):
        if not self.request.body:
            req_channels = channelstream.CHANNELS.keys()
            info_config = {
                'include_history': True,
                'include_users': True,
                'exclude_channels': [],
                'include_connections': True
            }
        else:
            json_body = self.request.json_body
            # get info config for channel information
            info_config = json_body.get('info') or {}
            req_channels = info_config.get('channels', None)
            info_config['include_connections'] = info_config.get(
                'include_connections', True)
        channels_info = self.get_common_info(req_channels, info_config)
        return channels_info



from gevent import monkey

monkey.patch_all()

from six.moves import configparser
import logging
import argparse

from gevent.server import StreamServer
from pyramid.settings import asbool
from channelstream.gc import gc_conns_forever, gc_users_forever
from channelstream.policy_server import client_handle
import channelstream.wsgi_app as pyramid_app
from channelstream.ws_app import ChatApplicationSocket

from ws4py.server.geventserver import WebSocketWSGIApplication, \
    WebSocketWSGIHandler, WSGIServer

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class RoutingApplication(object):
    def __init__(self, config):
        self.ws_app = WebSocketWSGIApplication(
            handler_cls=ChatApplicationSocket)
        self.wsgi_app = pyramid_app.make_app(config)

    def __call__(self, environ, start_response):
        if environ['PATH_INFO'] == '/ws':
            environ['ws4py.app'] = self
            return self.ws_app(environ, start_response)

        return self.wsgi_app(environ, start_response)


def cli_start():
    config = {
        'secret': '',
        'admin_secret': '',
        'gc_conns_after': 30,
        'gc_channels_after': 3600 * 72,
        'wake_connections_after': 5,
        'allow_posting_from': [],
        'port': 8000,
        'host': '0.0.0.0',
        'debug': False
    }

    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("-?", "--help", action="help")
    parser.add_argument("-i", "--ini", dest="ini",
                        help="config file location",
                        default=None)
    parser.add_argument("-s", "--secret", dest="secret",
                        help="secret used to secure your requests",
                        default='secret')
    parser.add_argument("-a", "--admin_secret", dest="admin_secret",
                        help="secret used to secure your admin_panel",
                        default='admin_secret')
    parser.add_argument("-host", "--host", dest="host",
                        help="host ip on which the server listens to",
                        default='0.0.0.0')
    parser.add_argument("-p", "--port", type=int, dest="port",
                        help="port on which the server listens to",
                        default=8000)
    parser.add_argument("-d", "--debug", dest="debug",
                        help="debug",
                        default=0)
    parser.add_argument("-e", "--demo", dest="demo",
                        help="demo enabled",
                        default=False)
    parser.add_argument("-x", "--allowed_post_ip", dest="allow_posting_from",
                        help="comma separated list of ip's "
                             "that can post to server",
                        default="127.0.0.1"
                        )
    args = parser.parse_args()
    if args.ini:
        parser = configparser.ConfigParser()
        parser.read(args.ini)

        non_optional_parameters = (
            'debug', 'port', 'host', 'secret', 'admin_secret',
            'demo_app_url', 'demo')
        for key in non_optional_parameters:
            try:
                config[key] = parser.get('channelstream', key)
            except configparser.NoOptionError:
                pass

        try:
            ips = [ip.strip() for ip in parser.get(
                'channelstream', 'allow_posting_from').split(',')]
            config['allow_posting_from'].extend(ips)
        except configparser.NoOptionError:
            pass

    else:
        config['debug'] = args.debug
        config['port'] = args.port
        config['demo'] = asbool(args.demo)
        config['host'] = args.host
        config['secret'] = args.secret
        config['admin_secret'] = args.admin_secret
        config['allow_posting_from'].extend(
            [ip.strip() for ip in args.allow_posting_from.split(',')])

    # convert types
    config['debug'] = int(config['debug'])
    config['port'] = int(config['port'])
    config['demo'] = asbool(args.demo)

    log_level = logging.DEBUG if config['debug'] else logging.INFO
    logging.basicConfig(level=log_level)

    url = 'http://{}:{}'.format(config['host'], config['port'])

    log.info('Starting flash policy server on port 10843')
    gc_conns_forever()
    gc_users_forever()
    server = StreamServer(('0.0.0.0', 10843), client_handle)
    server.start()
    log.info('Serving on {}'.format(url))
    log.info('Admin interface available on {}/admin'.format(url))
    if config['demo']:
        log.info('Demo enabled, visit {}/demo'.format(url))

    server = WSGIServer((config['host'], config['port']),
                        RoutingApplication(config))
    server.serve_forever()



import copy
import logging
import six
import uuid

from datetime import datetime

import channelstream

log = logging.getLogger(__name__)


class Channel(object):
    """ Represents one of our chat channels - has some config options """

    def __init__(self, name, long_name=None, channel_configs=None):
        self.name = name
        self.long_name = long_name
        self.last_active = datetime.utcnow()
        self.connections = {}
        self.notify_presence = False
        self.broadcast_presence_with_user_lists = False
        self.salvageable = False
        self.store_history = False
        self.history_size = 10
        self.history = []
        if not channel_configs:
            channel_configs = {}
        self.reconfigure_from_dict(channel_configs.get(self.name))
        log.info('%s created' % self)

    def reconfigure_from_dict(self, config):
        if config:
            keys = ('notify_presence', 'store_history', 'history_size',
                    'broadcast_presence_with_user_lists')
            for key in keys:
                val = config.get(key)
                if val is not None:
                    setattr(self, key, val)

    def add_connection(self, connection):
        username = connection.username
        if username not in self.connections:
            self.connections[username] = []

        if not self.connections[username] and self.notify_presence:
            self.send_notify_presence_info(username, 'joined')

        if connection not in self.connections[connection.username]:
            self.connections[connection.username].append(connection)

    def remove_connection(self, connection):
        username = connection.username
        if username not in self.connections:
            self.connections[username] = []

        if connection in self.connections[username]:
            self.connections[username].remove(connection)

        self.after_parted(username)

    def after_parted(self, username):
        """
        Sends parted message if necessary and removed username from
        connections if empty

        :param username:
        :return:
        """
        if not self.connections[username]:
            del self.connections[username]
            if self.notify_presence:
                self.send_notify_presence_info(username, 'parted')

    def send_notify_presence_info(self, username, action):
        """
        Sends a message to other connected parties about a presence change
        :param username:
        :param action:
        :return:
        """
        connected_users = []
        if self.broadcast_presence_with_user_lists:
            for _username in self.connections.keys():
                user_inst = channelstream.USERS.get(_username)
                user_data = {
                    'user': user_inst.username,
                    'state': user_inst.public_state
                }
                connected_users.append(user_data)

        self.last_active = datetime.utcnow()
        payload = {
            'uuid': str(uuid.uuid4()).replace('-', ''),
            'type': 'presence',
            'user': username,
            'users': connected_users,
            'timestamp': self.last_active,
            'channel': self.name,
            'message': {'action': action}
        }
        self.add_message(payload, exclude_users=username)
        return payload

    def add_message(self, message, pm_users=None, exclude_users=None):
        """
        Sends the message to all connections subscribed to this channel
        """
        message = copy.deepcopy(message)
        pm_users = pm_users or []
        self.last_active = datetime.utcnow()
        if self.store_history and message['type'] == 'message':
            self.history.append(message)
            self.history = self.history[(self.history_size) * -1:]
        message.update({'channel': self.name})
        # message everyone subscribed except excluded
        total_sent = 0
        for user, conns in six.iteritems(self.connections):
            if not exclude_users or user not in exclude_users:
                for connection in conns:
                    if not pm_users or connection.user in pm_users:
                        connection.add_message(message)
                        total_sent += 1
        return total_sent

    def __repr__(self):
        return '<Channel: %s, connections:%s>' % (
            self.name, len(self.connections))



"""
ex-json borrowed from Marcin Kuzminski

source: https://secure.rhodecode.org/ext-json

"""
import datetime
import functools
import decimal
import imp

__all__ = ['json', 'simplejson', 'stdlibjson']


def _is_aware(value):
    """
    Determines if a given datetime.time is aware.

    The logic is described in Python's docs:
    http://docs.python.org/library/datetime.html#datetime.tzinfo
    """
    return (value.tzinfo is not None
            and value.tzinfo.utcoffset(value) is not None)


def _obj_dump(obj):
    """
    Custom function for dumping objects to JSON, if obj has __json__ attribute
    or method defined it will be used for serialization

    :param obj:
    """

    if isinstance(obj, complex):
        return [obj.real, obj.imag]
    # See "Date Time String Format" in the ECMA-262 specification.
    # some code borrowed from django 1.4
    elif isinstance(obj, datetime.datetime):
        r = obj.isoformat()
        if obj.microsecond:
            r = r[:23] + r[26:]
        if r.endswith('+00:00'):
            r = r[:-6] + 'Z'
        return r
    elif isinstance(obj, datetime.date):
        return obj.isoformat()
    elif isinstance(obj, decimal.Decimal):
        return str(obj)
    elif isinstance(obj, datetime.time):
        if _is_aware(obj):
            raise ValueError("JSON can't represent timezone-aware times.")
        r = obj.isoformat()
        if obj.microsecond:
            r = r[:12]
        return r
    elif isinstance(obj, set):
        return list(obj)
    elif hasattr(obj, '__json__'):
        if callable(obj.__json__):
            return obj.__json__()
        else:
            return obj.__json__
    else:
        raise NotImplementedError


# Import simplejson
try:
    # import simplejson initially
    _sj = imp.load_module('_sj', *imp.find_module('simplejson'))

    def extended_encode(obj):
        try:
            return _obj_dump(obj)
        except NotImplementedError:
            pass
        raise TypeError("%r is not JSON serializable" % (obj,))

    # we handle decimals our own it makes unified behavior of json vs
    # simplejson
    sj_version = [int(x) for x in _sj.__version__.split('.')]
    major, minor = sj_version[0], sj_version[1]
    if major < 2 or (major == 2 and minor < 1):
        # simplejson < 2.1 doesnt support use_decimal
        _sj.dumps = functools.partial(_sj.dumps,
                                      default=extended_encode)
        _sj.dump = functools.partial(_sj.dump,
                                     default=extended_encode)
    else:
        _sj.dumps = functools.partial(_sj.dumps,
                                      default=extended_encode,
                                      use_decimal=False)
        _sj.dump = functools.partial(_sj.dump,
                                     default=extended_encode,
                                     use_decimal=False)
    simplejson = _sj

except ImportError:
    # no simplejson set it to None
    simplejson = None

try:
    # simplejson not found try out regular json module
    _json = imp.load_module('_json', *imp.find_module('json'))

    # extended JSON encoder for json
    class ExtendedEncoder(_json.JSONEncoder):
        def default(self, obj):
            try:
                return _obj_dump(obj)
            except NotImplementedError:
                pass
            raise TypeError("%r is not JSON serializable" % (obj,))

    # monkey-patch JSON encoder to use extended version
    _json.dumps = functools.partial(_json.dumps, cls=ExtendedEncoder)
    _json.dump = functools.partial(_json.dump, cls=ExtendedEncoder)

except ImportError:
    json = None

stdlibjson = _json

# set all available json modules
if simplejson:
    json = _sj
elif _json:
    json = _json
else:
    raise ImportError('Could not find any json modules')



from six.moves.urllib.parse import parse_qs
from datetime import datetime

from ws4py.websocket import WebSocket

import channelstream


class ChatApplicationSocket(WebSocket):
    def opened(self):
        self.qs = parse_qs(self.environ['QUERY_STRING'])
        self.conn_id = self.qs['conn_id'][0]
        if self.conn_id not in channelstream.CONNECTIONS:
            # close connection instantly if user played with id
            self.close()
        else:
            # attach a socket to connection
            channelstream.CONNECTIONS[self.conn_id].socket = self

    def received_message(self, m):
        # this is to allow client heartbeats
        now = datetime.utcnow()
        if self.conn_id in channelstream.CONNECTIONS:
            connection = channelstream.CONNECTIONS[self.conn_id]
            connection.last_active = now
            user = channelstream.USERS.get(connection.username)
            if user:
                user.last_active = now

    def closed(self, code, reason=""):
        self.environ.pop('ws4py.app')
        found_conn = self.conn_id in channelstream.CONNECTIONS
        if hasattr(self, 'conn_id') and found_conn:
            connection = channelstream.CONNECTIONS[self.conn_id]
            connection.mark_for_gc()



from datetime import datetime
from gevent.lock import RLock

__version__ = {
    'major': 0,
    'minor': 5,
    'patch': 1
}

stats = {
    'total_messages': 0,
    'total_unique_messages': 0,
    'started_on': datetime.utcnow()
}

lock = RLock()
CHANNELS = {}
USERS = {}
CONNECTIONS = {}



import logging

from datetime import datetime


log = logging.getLogger(__name__)


class User(object):
    """ represents a unique user of the system """

    def __init__(self, username):
        self.username = username
        self.state = {}
        self.state_public_keys = []
        self.connections = []  # holds ids of connections
        self.last_active = datetime.utcnow()

    def __repr__(self):
        return '<User:%s, connections:%s>' % (
            self.username, len(self.connections))

    def add_connection(self, connection):
        """
        creates a new connection for user
        """
        if connection not in self.connections:
            self.connections.append(connection)
        # mark active
        self.last_active = datetime.utcnow()
        return connection

    def add_message(self, message):
        """
        Send a message to all connections of this user
        """
        # mark active
        self.last_active = datetime.utcnow()
        for connection in self.connections:
            connection.add_message(message)
        return len(self.connections)

    def state_from_dict(self, state_dict):
        if isinstance(state_dict, dict):
            self.state.update(state_dict)

    @property
    def public_state(self):
        return {k: v for k, v in self.state.items()
                if k in self.state_public_keys}



"""View handlers package.
"""
import logging


log = logging.getLogger(__name__)


def includeme(config):
    config.add_route('index', '/')
    config.add_route('demo', '/demo')
    config.add_route('admin', '/admin',
                     factory='channelstream.wsgi_views.'
                             'wsgi_security:BasicAuthFactory')
    config.add_route('action', '/{action}')
    config.add_route('section_action', '/{section}/{action}')



import json
import random
import uuid
import requests
import six
from pyramid.view import view_config
from itsdangerous import TimestampSigner

POSSIBLE_CHANNELS = set(['pub_chan', 'pub_chan2', 'notify'])


def make_request(request, payload, endpoint):
    server_port = request.registry.settings['port']
    signer = TimestampSigner(request.registry.settings['secret'])
    sig_for_server = signer.sign(endpoint)
    if not six.PY2:
        sig_for_server = sig_for_server.decode('utf8')
    secret_headers = {'x-channelstream-secret': sig_for_server,
                      'x-channelstream-endpoint': endpoint,
                      'Content-Type': 'application/json'}
    url = 'http://127.0.0.1:%s%s' % (server_port, endpoint)
    response = requests.post(url, data=json.dumps(payload),
                             headers=secret_headers)
    return response.json()


def enable_demo(context, request):
    if request.registry.settings['demo']:
        return True
    return False


CHANNEL_CONFIGS = {"pub_chan": {"notify_presence": True},
                   "notify": {"notify_presence": True}}


class DemoViews(object):
    def __init__(self, request):
        self.request = request
        self.request.response.headers.add('Cache-Control',
                                          'no-cache, no-store')

    @view_config(route_name='section_action', renderer='string',
                 request_method="OPTIONS", custom_predicates=[enable_demo])
    def handle_CORS(self):
        self.request.response.headers.add('Access-Control-Allow-Origin', '*')
        self.request.response.headers.add('XDomainRequestAllowed', '1')
        self.request.response.headers.add('Access-Control-Allow-Methods',
                                          'GET, POST, OPTIONS, PUT')
        self.request.response.headers.add('Access-Control-Allow-Headers',
                                          'Content-Type, Depth, User-Agent, '
                                          'X-File-Size, X-Requested-With, '
                                          'If-Modified-Since, X-File-Name, '
                                          'Cache-Control, Pragma, Origin, '
                                          'Connection, Referer, Cookie')
        self.request.response.headers.add('Access-Control-Max-Age', '86400')
        # self.request.response.headers.add('Access-Control-Allow-Credentials',
        # 'true')
        return {}

    @view_config(route_name='demo', renderer='templates/demo.jinja2',
                 custom_predicates=[enable_demo])
    def demo(self):
        random_name = 'anon_%s' % random.randint(1, 999999)
        return {'username': random_name}

    @view_config(route_name='section_action',
                 match_param=['section=demo', 'action=connect'],
                 renderer='json', request_method="POST",
                 custom_predicates=[enable_demo])
    def connect(self):
        """handle authorization of users trying to connect"""
        channels = self.request.json_body['channels']
        POSSIBLE_CHANNELS.intersection(channels)
        random_name = 'anon_%s' % random.randint(1, 999999)
        username = self.request.json_body.get('username', random_name)

        payload = {"username": username,
                   "conn_id": str(uuid.uuid4()),
                   "channels": channels,
                   "channel_configs": CHANNEL_CONFIGS}
        make_request(self.request, payload, '/connect')
        return payload

    @view_config(route_name='section_action',
                 match_param=['section=demo', 'action=subscribe'],
                 renderer='json', request_method="POST",
                 custom_predicates=[enable_demo])
    def subscribe(self):
        """"can be used to subscribe specific connection to other channels"""
        request_data = self.request.json_body
        channels = request_data['channels']
        POSSIBLE_CHANNELS.intersection(channels)
        payload = {"conn_id": request_data.get('conn_id', ''),
                   "channels": request_data.get('channels', []),
                   "channel_configs": CHANNEL_CONFIGS
                   }
        result = make_request(self.request, payload, '/subscribe')
        return result['channels']

    @view_config(route_name='section_action',
                 match_param=['section=demo', 'action=unsubscribe'],
                 renderer='json', request_method="POST",
                 custom_predicates=[enable_demo])
    def unsubscribe(self):
        """"can be used to subscribe specific connection to other channels"""
        request_data = self.request.json_body
        payload = {"conn_id": request_data.get('conn_id', ''),
                   "channels": request_data.get('channels', [])
                   }
        result = make_request(self.request, payload, '/unsubscribe')
        return result['channels']

    @view_config(route_name='section_action',
                 match_param=['section=demo', 'action=message'],
                 renderer='json', request_method="POST",
                 custom_predicates=[enable_demo])
    def message(self):
        """send message to channel/users"""
        request_data = self.request.json_body
        payload = {
            'type': 'message',
            "user": request_data.get('user', 'system'),
            "channel": request_data.get('channel', 'unknown_channel'),
            'message': request_data.get('message')
        }
        result = make_request(self.request, [payload], '/message')
        return result

    @view_config(route_name='section_action',
                 match_param=['section=demo', 'action=channel_config'],
                 renderer='json', request_method="POST",
                 custom_predicates=[enable_demo])
    def channel_config(self):
        """configure channel defaults"""

        payload = [
            ('pub_chan', {
                "notify_presence": True,
                "store_history": True,
                "history_size": 20
            }),
            ('pub_chan2', {
                "notify_presence": True,
                "salvageable": True,
                "store_history": True,
                "history_size": 30
            })
        ]
        url = 'http://127.0.0.1:%s/channel_config' % self.server_port
        response = requests.post(url, data=json.dumps(payload),
                                 headers=self.secret_headers).json()
        return response



import os
import re
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
REQUIREMENTS = open(os.path.join(here, 'requirements.txt')).readlines()

compiled = re.compile('([^=><]*).*')

def parse_req(req):
    return compiled.search(req).group(1).strip()


requires = [_f for _f in map(parse_req, REQUIREMENTS) if _f]

setup(name='channelstream',
      version='0.5.1',
      description='Websocket server supporting channels/users communication',
      classifiers=[
          'Intended Audience :: Developers'
      ],
      author='Marcin Lulek',
      author_email='info@webreactor.eu',
      license='BSD',
      zip_safe=True,
      packages=find_packages(),
      include_package_data=True,
      install_requires=requires,
      entry_points={
          'console_scripts': [
              'channelstream = channelstream.cli:cli_start',
          ],
      })



import logging
from datetime import datetime, timedelta

import gevent

import channelstream
from .ext_json import json

log = logging.getLogger(__name__)


class Connection(object):
    """ Represents a browser connection"""

    def __init__(self, username, conn_id):
        self.username = username  # hold user id/name of connection
        self.last_active = datetime.utcnow()
        self.socket = None
        self.queue = None
        self.id = conn_id
        gevent.spawn_later(5, self.heartbeat)

    def __repr__(self):
        return '<Connection: id:%s, owner:%s>' % (self.id, self.username)

    def add_message(self, message=None):
        """ Sends the message to the client connection """
        # handle websockets
        if self.socket and self.socket.terminated:
            self.mark_for_gc()
        elif self.socket and not self.socket.terminated:
            try:
                # payload needs to be converted to JSON now as it gets
                # piped to client
                self.socket.send(json.dumps([message] if message else []))
                now = datetime.utcnow()
                self.last_active = now
                channelstream.USERS[self.username].last_active = now
            except Exception:
                self.mark_for_gc()
        elif self.queue:
            # handle long polling
            # payload will be converted to JSON in WSGI response
            self.queue.put([message] if message else [])

    def mark_for_gc(self):
        # set last active time for connection 1 hour in past for GC
        self.last_active -= timedelta(days=60)

    def heartbeat(self):
        if self.socket or self.queue:
            try:
                self.add_message()
                gevent.spawn_later(5, self.heartbeat)
            except Exception:
                self.mark_for_gc()
                if self.socket:
                    self.socket.close()



import socket
import logging

log = logging.getLogger(__name__)

POLICY = '<cross-domain-policy><allow-access-from domain="*" ' \
         'to-ports="*" /></cross-domain-policy>\0'
POLICYREQUEST = "<policy-file-request/>"


def client_handle(sock, address):
    log.info("%s:%s: Connection accepted." % address)
    sock.settimeout(3)
    try:
        input = sock.recv(128)
        if input.startswith(POLICYREQUEST):
            sock.sendall(POLICY)
            log.info("%s:%s: Policy sent. Closing connection." % address)
        else:
            log.info("%s:%s: Wrong payload. Closing connection." % address)
    except socket.timeout:
        log.info("%s:%s: Timeout" % address)
    sock.close()



from gevent import monkey

monkey.patch_all()

import pytest
from datetime import datetime, timedelta
from gevent.queue import Queue, Empty
from pyramid import testing
import channelstream
import channelstream.gc
from channelstream.channel import Channel
from channelstream.connection import Connection
from channelstream.user import User


@pytest.fixture
def cleanup_globals():
    channelstream.CHANNELS = {}
    channelstream.CONNECTIONS = {}
    channelstream.USERS = {}


@pytest.mark.usefixtures("cleanup_globals")
class TestChannel(object):
    def test_create_defaults(self):
        channel = Channel('test', long_name='long name')
        assert channel.name == 'test'
        assert channel.long_name == 'long name'
        assert channel.connections == {}
        assert channel.notify_presence is False
        assert channel.broadcast_presence_with_user_lists is False
        assert channel.salvageable is False
        assert channel.store_history is False
        assert channel.history_size == 10
        assert channel.history == []

    def test_repr(self):
        channel = Channel('test', long_name='long name')
        assert repr(channel) == '<Channel: test, connections:0>'

    @pytest.mark.parametrize('prop, value', [
        ('notify_presence', True),
        ('store_history', 6),
        ('history_size', 42),
        ('broadcast_presence_with_user_lists', True)
    ])
    def test_create_set_config(self, prop, value):
        channel_configs = {'test': {prop: value}}
        channel = Channel('test', channel_configs=channel_configs)

        assert getattr(channel, prop) == value

    def test_create_set_config_diff_name(self):
        channel_configs = {'test2': {'notify_presence': True}}
        channel = Channel('test', channel_configs=channel_configs)
        assert channel.notify_presence is False

    def test_add_connection(self):
        connection = Connection('test_user',
                                conn_id='A')
        channel = Channel('test')
        channel.add_connection(connection)
        assert len(channel.connections['test_user']) == 1
        assert 'test_user' in channel.connections
        assert connection in channel.connections['test_user']
        assert repr(channel) == '<Channel: test, connections:1>'

    def test_remove_connection(self):
        connection = Connection('test_user', conn_id='A')
        connection2 = Connection('test_user2', conn_id='B')
        connection3 = Connection('test_user', conn_id='C')
        channel = Channel('test')
        channel.add_connection(connection)
        channel.add_connection(connection2)
        channel.remove_connection(connection)
        assert 'test_user' not in channel.connections
        assert len(channel.connections['test_user2']) == 1
        channel.add_connection(connection)
        channel.add_connection(connection3)
        channel.remove_connection(connection)
        assert len(channel.connections['test_user']) == 1

    def test_remove_non_existant_connection(self):
        channel = Channel('test')
        connection = Connection('test_user', conn_id='A')
        channel.remove_connection(connection)
        assert 'test_user' not in channel.connections

    def test_remove_connection_w_presence(self):
        user = User('test_user')
        channelstream.USERS[user.username] = user
        connection = Connection('test_user', conn_id='A')
        user.add_connection(connection)
        config = {'test': {'notify_presence': True,
                           'broadcast_presence_with_user_lists': True}}
        channel = Channel('test', channel_configs=config)
        channel.add_connection(connection)
        channel.remove_connection(connection)

    def test_add_connection_w_presence(self):
        user = User('test_user')
        channelstream.USERS[user.username] = user
        connection = Connection('test_user', conn_id='A')
        user.add_connection(connection)
        config = {'test': {'notify_presence': True,
                           'broadcast_presence_with_user_lists': True}}
        channel = Channel('test', channel_configs=config)
        channel.add_connection(connection)
        assert len(channel.connections['test_user']) == 1
        assert 'test_user' in channel.connections
        assert connection in channel.connections['test_user']

    def test_presence_message(self):
        user = User('test_user')
        connection = Connection('test_user', conn_id='A')
        user.add_connection(connection)
        channel = Channel('test')
        channel.add_connection(connection)
        payload = channel.send_notify_presence_info('test_user', 'join')
        assert payload['user'] == 'test_user'
        assert payload['message']['action'] == 'join'
        assert payload['type'] == 'presence'
        assert payload['channel'] == 'test'
        assert len(payload['users']) == 0

    def test_presence_message_w_users(self):
        user = User('test_user')
        user.state_from_dict({'key': '1', 'key2': '2'})
        user.state_public_keys = ['key2']
        channelstream.USERS[user.username] = user
        connection = Connection('test_user', conn_id='A')
        user.add_connection(connection)
        user2 = User('test_user2')
        user2.state_from_dict({'key': '1', 'key2': '2'})
        channelstream.USERS[user2.username] = user2
        connection2 = Connection('test_user2', conn_id='A')
        user2.add_connection(connection2)
        config = {'test': {'notify_presence': True,
                           'broadcast_presence_with_user_lists': True}}
        channel = Channel('test', channel_configs=config)
        channel.add_connection(connection)
        channel.add_connection(connection2)
        payload = channel.send_notify_presence_info('test_user', 'join')
        assert len(payload['users']) == 2
        sorted_users = sorted(payload['users'], key=lambda x: x['user'])
        assert sorted_users == [
            {'state': {'key2': '2'}, 'user': 'test_user'},
            {'state': {}, 'user': 'test_user2'}
        ]

    def test_history(self):
        config = {'test': {'store_history': True,
                           'history_size': 3}}
        channel = Channel('test', long_name='long name',
                          channel_configs=config)
        channel.add_message({'message': 'test1', 'type': 'message'})
        channel.add_message({'message': 'test2', 'type': 'message'})
        channel.add_message({'message': 'test3', 'type': 'message'})
        channel.add_message({'message': 'test4', 'type': 'message'})

        assert len(channel.history) == 3
        assert channel.history == [
            {'channel': 'test', 'message': 'test2', 'type': 'message'},
            {'channel': 'test', 'message': 'test3', 'type': 'message'},
            {'channel': 'test', 'message': 'test4', 'type': 'message'}
        ]


@pytest.mark.usefixtures("cleanup_globals")
class TestConnection(object):
    def test_create_defaults(self):
        now = datetime.utcnow()
        connection = Connection('test', 'X')
        assert connection.username == 'test'
        assert now <= connection.last_active
        assert connection.socket is None
        assert connection.queue is None
        assert connection.id == 'X'

    def test_mark_for_gc(self):
        long_time_ago = datetime.utcnow() - timedelta(days=50)
        connection = Connection('test', 'X')
        connection.mark_for_gc()
        assert connection.last_active < long_time_ago

    def test_message(self):
        connection = Connection('test', 'X')
        connection.queue = Queue()
        connection.add_message({'message': 'test'})
        assert connection.queue.get() == [{'message': 'test'}]

    def test_heartbeat(self):
        connection = Connection('test', 'X')
        connection.queue = Queue()
        connection.heartbeat()
        assert connection.queue.get() == []


class TestUser(object):
    def test_create_defaults(self):
        user = User('test_user')
        user.state_from_dict({'key': '1', 'key2': '2'})
        user.state_public_keys = ['key2']
        assert repr(user) == '<User:test_user, connections:0>'
        assert sorted(user.state.items()) == sorted({'key': '1',
                                                     'key2': '2'}.items())
        assert user.public_state == {'key2': '2'}

    def test_messages(self):
        user = User('test_user')
        connection = Connection('test_user', conn_id='A')
        connection.queue = Queue()
        connection2 = Connection('test_user', conn_id='B')
        connection2.queue = Queue()
        user.add_connection(connection)
        user.add_connection(connection2)
        user.add_message({'type': 'message'})
        assert len(user.connections) == 2
        assert len(user.connections[0].queue.get()) == 1
        assert len(user.connections[1].queue.get()) == 1


@pytest.mark.usefixtures("cleanup_globals")
class TestGC(object):
    def test_gc_connections_active(self):
        channel = Channel('test')
        channelstream.CHANNELS[channel.name] = channel
        channel2 = Channel('test2')
        channelstream.CHANNELS[channel2.name] = channel2
        user = User('test_user')
        channelstream.USERS[user.username] = user
        user2 = User('test_user2')
        channelstream.USERS[user2.username] = user2
        connection = Connection('test_user', '1')
        channelstream.CONNECTIONS[connection.id] = connection
        connection2 = Connection('test_user', '2')
        channelstream.CONNECTIONS[connection2.id] = connection2
        connection3 = Connection('test_user2', '3')
        channelstream.CONNECTIONS[connection3.id] = connection3
        connection4 = Connection('test_user2', '4')
        channelstream.CONNECTIONS[connection4.id] = connection4
        user.add_connection(connection)
        user.add_connection(connection2)
        channel.add_connection(connection)
        channel.add_connection(connection2)
        user2.add_connection(connection3)
        user2.add_connection(connection4)
        channel2.add_connection(connection3)
        channel2.add_connection(connection4)
        channelstream.gc.gc_conns()
        conns = channelstream.CHANNELS['test'].connections['test_user']
        assert len(conns) == 2
        assert len(channelstream.CONNECTIONS.items()) == 4
        conns = channelstream.CHANNELS['test2'].connections['test_user2']
        assert len(conns) == 2
        assert len(user.connections) == 2
        assert len(user2.connections) == 2
        assert sorted(channel.connections.keys()) == ['test_user']
        assert sorted(channel2.connections.keys()) == ['test_user2']

    def test_gc_connections_collecting(self):
        channel = Channel('test')
        channelstream.CHANNELS[channel.name] = channel
        channel2 = Channel('test2')
        channelstream.CHANNELS[channel2.name] = channel2
        user = User('test_user')
        channelstream.USERS[user.username] = user
        user2 = User('test_user2')
        channelstream.USERS[user2.username] = user2
        connection = Connection('test_user', '1')
        channelstream.CONNECTIONS[connection.id] = connection
        connection2 = Connection('test_user', '2')
        connection2.mark_for_gc()
        channelstream.CONNECTIONS[connection2.id] = connection2
        connection3 = Connection('test_user2', '3')
        connection3.mark_for_gc()
        channelstream.CONNECTIONS[connection3.id] = connection3
        connection4 = Connection('test_user2', '4')
        channelstream.CONNECTIONS[connection4.id] = connection4
        user.add_connection(connection)
        user.add_connection(connection2)
        channel.add_connection(connection)
        channel.add_connection(connection2)
        user2.add_connection(connection3)
        user2.add_connection(connection4)
        channel2.add_connection(connection3)
        channel2.add_connection(connection4)
        channelstream.gc.gc_conns()
        assert len(channelstream.CONNECTIONS.items()) == 2
        conns = channelstream.CHANNELS['test'].connections['test_user']
        assert len(conns) == 1
        assert conns == [connection]
        conns = channelstream.CHANNELS['test2'].connections['test_user2']
        assert len(conns) == 1
        assert conns == [connection4]
        assert len(user.connections) == 1
        assert len(user2.connections) == 1
        connection.mark_for_gc()
        connection4.mark_for_gc()
        channelstream.gc.gc_conns()
        assert 'test_user' not in channelstream.CHANNELS['test'].connections
        assert 'test_user2' not in channelstream.CHANNELS['test2'].connections
        assert len(channelstream.CHANNELS['test'].connections.items()) == 0
        assert len(channelstream.CHANNELS['test2'].connections.items()) == 0

    def test_users_active(self):
        user = User('test_user')
        channelstream.USERS[user.username] = user
        user2 = User('test_user2')
        channelstream.USERS[user2.username] = user2
        channelstream.gc.gc_users()
        assert len(channelstream.USERS.items()) == 2
        user.last_active -= timedelta(days=2)
        channelstream.gc.gc_users()
        assert len(channelstream.USERS.items()) == 1


@pytest.fixture
def pyramid_config():
    # from pyramid.request import Request
    # request = Request.blank('/', base_url='http://foo.com')
    config = testing.setUp(settings={})
    settings = config.get_settings()
    return config, settings


@pytest.fixture
def dummy_request():
    return testing.DummyRequest()


@pytest.mark.usefixtures('cleanup_globals', 'pyramid_config')
class TestConnectViews(object):
    def test_bad_json(self, dummy_request):
        from channelstream.wsgi_views.server import ServerViews
        dummy_request.json_body = {}
        view_cls = ServerViews(dummy_request)
        result = view_cls.connect()
        assert result == {'error': 'No username specified'}

    def test_good_json(self, dummy_request):
        from channelstream.wsgi_views.server import ServerViews
        dummy_request.json_body = {'username': 'username',
                                   'conn_id': 'X',
                                   'fresh_user_state': {'key': 'foo'},
                                   'user_state': {'bar': 'baz'},
                                   'state_public_keys': 'bar',
                                   'channels': ['a', 'aB'],
                                   'channel_configs': {
                                       'a': {'store_history': True,
                                             'history_size': 2}}}
        view_cls = ServerViews(dummy_request)
        assert channelstream.CHANNELS == {}
        result = view_cls.connect()
        assert len(channelstream.CHANNELS.keys()) == 2
        assert 'username' in channelstream.USERS
        assert 'X' in channelstream.CONNECTIONS
        assert result['channels'] == ['a', 'aB']
        assert result['state'] == {'bar': 'baz', 'key': 'foo'}
        assert result['conn_id'] == 'X'
        channels_info = result['channels_info']['channels']
        assert len(channels_info.keys()) == 2
        assert result['channels_info']['unique_users'] == 1
        assert channels_info['a']['total_users'] == 1
        assert channels_info['a']['total_connections'] == 1
        assert channels_info['a']['users'] == [{'connections': [],
                                                'user': 'username'}]
        assert channels_info['a']['history'] == []
        assert result['channels_info']['users'] == [
            {'state': {'bar': 'baz', 'key': 'foo'}, 'user': 'username'}
        ]


@pytest.mark.usefixtures('cleanup_globals', 'pyramid_config')
class TestSubscribeViews(object):
    def test_bad_json(self, dummy_request):
        from channelstream.wsgi_views.server import ServerViews
        dummy_request.json_body = {}
        view_cls = ServerViews(dummy_request)
        result = view_cls.subscribe()
        assert result == {'error': 'Unknown connection'}

    def test_good_json(self, dummy_request):
        from channelstream.wsgi_views.server import ServerViews
        dummy_request.json_body = {'username': 'test',
                                   'conn_id': 'x',
                                   'fresh_user_state': {'key': 'foo'},
                                   'user_state': {'bar': 'baz'},
                                   'state_public_keys': 'bar',
                                   'channels': ['a', 'aB'],
                                   'channel_configs': {
                                       'a': {'store_history': True,
                                             'history_size': 2}}}
        view_cls = ServerViews(dummy_request)
        view_cls.connect()
        dummy_request.json_body = {"conn_id": 'x',
                                   "channels": ['b'],
                                   "channel_configs": {
                                       "a": {"notify_presence": True},
                                       "b": {"notify_presence": True}}
                                   }
        view_cls = ServerViews(dummy_request)
        result = view_cls.subscribe()
        assert sorted(result['channels']) == sorted(['a', 'aB', 'b'])
        assert result['channels_info']['users'] == [
            {'state': {'bar': 'baz', 'key': 'foo'}, 'user': 'test'}]
        assert result['channels_info']['unique_users'] == 1
        assert 'a' in result['channels_info']['channels']
        assert 'b' in result['channels_info']['channels']
        assert result['channels_info']['channels']['a'][
                   'total_connections'] == 1
        assert result['channels_info']['channels']['a']['total_users'] == 1
        assert result['channels_info']['channels']['a']['history'] == []
        assert result['channels_info']['channels']['a']['users'] == [
            {'connections': [], 'user': 'test'}
        ]


@pytest.mark.usefixtures('cleanup_globals', 'pyramid_config')
class TestInfoView(object):
    def test_empty_json(self, dummy_request):
        from channelstream.wsgi_views.server import ServerViews
        dummy_request.json_body = {}
        view_cls = ServerViews(dummy_request)
        result = view_cls.info()
        assert result['channels'] == {}
        assert result['unique_users'] == 0
        assert result['users'] == []

    def test_subscribed_json(self, dummy_request):
        from channelstream.wsgi_views.server import ServerViews
        dummy_request.json_body = {'username': 'test1',
                                   'conn_id': 'x',
                                   'fresh_user_state': {'key': 'foo'},
                                   'user_state': {'bar': 'baz'},
                                   'state_public_keys': 'bar',
                                   'channels': ['a', 'aB'],
                                   'channel_configs': {
                                       'a': {'store_history': True,
                                             'history_size': 2}}}
        view_cls = ServerViews(dummy_request)
        view_cls.connect()
        dummy_request.json_body = {'username': 'test2',
                                   'conn_id': 'y',
                                   'fresh_user_state': {'key': 'foo1'},
                                   'user_state': {'bar': 'baz1'},
                                   'state_public_keys': 'key',
                                   'channels': ['a', 'c'],
                                   'channel_configs': {
                                       'c': {'store_history': True,
                                             'history_size': 2}}}
        view_cls = ServerViews(dummy_request)
        view_cls.connect()
        result = view_cls.info()
        assert sorted(('a', 'aB', 'c')) == sorted(result['channels'].keys())
        assert result['users']
        compA = sorted(result['channels']['a']['users'],
                       key=lambda x: x['user'])
        compB = sorted([
            {'connections': ['x'], 'user': 'test1'},
            {'connections': ['y'], 'user': 'test2'}],
            key=lambda x: x['user'])
        print(compA, compB)
        assert compA == compB
        assert result['channels']['a']['total_users'] == 2
        assert result['channels']['a']['total_connections'] == 2
        assert result['channels']['c']['users'] == [
            {'connections': ['y'], 'user': 'test2'}
        ]
        assert result['channels']['c']['total_users'] == 1
        assert result['channels']['c']['total_connections'] == 1
        assert result['channels']['aB']['users'] == [
            {'connections': ['x'], 'user': 'test1'}
        ]
        compA = sorted(result['users'],
                       key=lambda x: x['user'])
        compB = sorted([
            {'state': {'bar': 'baz', 'key': 'foo'}, 'user': 'test1'},
            {'state': {'bar': 'baz1', 'key': 'foo1'}, 'user': 'test2'}],
            key=lambda x: x['user'])
        print(compA, compB)
        assert compA == compB
        dummy_request.body = 'NOTEMPTY'
        dummy_request.json_body = {'info': {'channels': ['a']}}
        view_cls = ServerViews(dummy_request)
        result = view_cls.info()
        assert 'a' in result['channels']
        assert 'aB' not in result['channels']
